/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 04/03/23.
//

#ifndef FGGL_GUI_MODEL_STRUCTURE_HPP
#define FGGL_GUI_MODEL_STRUCTURE_HPP

#include <map>
#include <variant>
#include <vector>
#include <string>

#include <fggl/math/types.hpp>

namespace fggl::gui::model {

	using AttrKey = std::string;
	using AttrValue = std::variant<bool, int, float, std::string, math::vec2, math::vec3>;

	class Widget {
		public:
			using ChildItr = std::vector<Widget>::iterator;
			using ChildItrConst = std::vector<Widget>::const_iterator;

			inline bool isRoot() const {
				return m_parent == nullptr;
			}

			inline bool isLeaf() const {
				return m_children.empty();
			}

			inline bool hasAttr(const AttrKey& key) const {
				return m_attrs.contains(key);
			}

			inline std::map<AttrKey, AttrValue> attrs() {
				return m_attrs;
			}

			template<typename type>
			inline type get(const AttrKey& key) const {
				return std::get<type>(m_attrs.at(key));
			}

			template<typename type>
			type get_or_default(const AttrKey& key) const {
				auto itr = m_attrs.find(key);
				if ( itr == m_attrs.end() ) {
					return type();
				} else {
					return std::get<type>(itr->second);
				}
			}

			template<typename type>
			inline void set(const AttrKey& key, type value) {
				m_attrs[key] = value;
			}

			inline void addChild(const Widget& element) {
				auto childItr = m_children.insert( m_children.cend(), element );
				childItr->m_parent = this;
			}

			inline ChildItr begin() {
				return m_children.begin();
			}

			inline ChildItrConst begin() const noexcept {
				return m_children.begin();
			}

			inline ChildItrConst cbegin() const noexcept {
				return m_children.cbegin();
			}

			inline ChildItr end() {
				return m_children.end();
			}

			inline ChildItrConst end() const {
				return m_children.end();
			}

			inline ChildItrConst cend() const {
				return m_children.cend();
			}

		private:
			Widget *m_parent;
			std::vector<Widget> m_children;
			std::map< AttrKey, AttrValue > m_attrs;
	};

	void attr_box_set(Widget& widget, const AttrKey& key, auto top, auto right, auto bottom, auto left) {
		widget.set(key + "::top", top);
		widget.set(key + "::right", right);
		widget.set(key + "::bottom", bottom);
		widget.set(key + "::left", left);
	}

	inline void attr_box_set(Widget& widget, const AttrKey& key, auto vert, auto horz) {
		attr_box_set(widget, key, vert, horz, vert, horz);
	}

	inline void attr_box_set(Widget& widget, const AttrKey& key, auto value) {
		attr_box_set(widget, key, value, value, value, value);
	}


} // namespace fggl::gui::model

#endif //FGGL_GUI_MODEL_STRUCTURE_HPP
